#include <LocoNet.h>
#include <EEPROMex.h>

#define ARTNR 5002

#define LNCV_COUNT 3

const int EEPROM_VERSION = 5;
const int EEPROM_KEY = ARTNR * 1000 + EEPROM_VERSION;

#define  TX_PIN   7

static   LnBuf        LnTxBuffer ;
static   lnMsg        *LnPacket;

const int BLOCK_COUNT = 16;
const int LED_PIN = 13;
const int BROADCAST_DELAY = 2000;//millis

#define CV_ADDR 0
#define CV_ON_DELAY 1
#define CV_REPLAY_MSG_DELAY 2

uint16_t lncv[LNCV_COUNT] = {10, 1000, 500};

volatile boolean programmingMode = 0;
volatile unsigned long programmingModeStart = 0;
unsigned long ledCycle = 500;
unsigned long PROG_MODE_TIMEOUT = 30000; //mills, 60sec

//for led indication
unsigned long LedLastBlink = 0;
byte          LedState = 0;

int Pins[BLOCK_COUNT] = {6,5,4,2,3,10,11,12,9,A6,A7,A5,A4,A3,A2,A1};

unsigned long LastOn[BLOCK_COUNT] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
unsigned long Tests[BLOCK_COUNT] =  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
unsigned long LastBroadcast;
unsigned long LastSend[BLOCK_COUNT] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};



LocoNetCVClass lnCV;

byte CurrentBlock;
byte CurrentBroadcastSensor;

void EEPROMSaveCV(int CV) {
  EEPROM.writeInt(sizeof(EEPROM_KEY) + sizeof(lncv[0]) * CV, lncv[CV]);
}

void setup(){
  Serial.begin(57600);
  
  int testKey ;
  EEPROM.readBlock(0,testKey);

  if (testKey != EEPROM_KEY) {
    Serial.print("FIRST START address:");
    EEPROM.writeBlock(0, EEPROM_KEY);
    EEPROM.writeBlock(sizeof(EEPROM_KEY),lncv);
  }
  else {
    EEPROM.readBlock(sizeof(EEPROM_KEY),lncv);
    Serial.print("loaded address:");
  }
  Serial.println(lncv[CV_ADDR]);
 
  // First initialize the LocoNet interface, specifying the TX Pin
  LocoNet.init(TX_PIN);
 
  // Initialize a LocoNet packet buffer to buffer bytes from the PC
  initLnBuf(&LnTxBuffer) ;

  pinMode(LED_PIN, OUTPUT);
  //digitalWrite(LED_PIN, HIGH);
  
  for (int i = 0; i < BLOCK_COUNT; i++) {
    pinMode(Pins[i], INPUT_PULLUP);
    LastOn[i] = 0;
    LastSend[i] = 0;
    Tests[i] = GetSensor(i);
  }    
  CurrentBlock = 0;
  LastBroadcast = 0;
  CurrentBroadcastSensor = 0;


  Serial.print("Track occupancy detectror ver:");
  Serial.print(EEPROM_VERSION);
  Serial.print(" addr:");
  Serial.println(lncv[CV_ADDR]);

  //sendAll();
  programmingMode = false;

}

bool GetSensor(int CurrentBlock){
  if (Pins[CurrentBlock] > 13) 
    return analogRead(Pins[CurrentBlock]) < 300 ? true : false;
  else 
    return !digitalRead(Pins[CurrentBlock]);
}

void loop(){
byte val;

 LnPacket = LocoNet.receive();
  if (LnPacket) {
    if ( LnPacket -> data[0] == 0x83 ) sendAll();

    uint8_t packetConsumed(LocoNet.processSwitchSensorMessage(LnPacket));
    if (packetConsumed == 0) {
      Serial.print("Loop ");
      Serial.print((int)LnPacket);
      dumpPacket(LnPacket->ub);
      packetConsumed = lnCV.processLNCVMessage(LnPacket);
      Serial.print("End Loop\n");
    }
  }


//read current sensor
  Tests[CurrentBlock] = Tests[CurrentBlock] << 1;
  
  val = GetSensor(CurrentBlock);
 // if (Pins[CurrentBlock] > 13) val = analogRead(Pins[CurrentBlock]) < 300 ? true : false;
 //   else val = !digitalRead(Pins[CurrentBlock]);
  bitWrite(Tests[CurrentBlock], 0, val);

//    
  if (LastOn[CurrentBlock]) {
    if (millis()-LastOn[CurrentBlock] > lncv[CV_ON_DELAY])
         if (!Tests[CurrentBlock]) BlockFree(CurrentBlock); else BlockBusy(CurrentBlock, millis());
  }
  else  
    if (Tests[CurrentBlock]) BlockBusy(CurrentBlock, millis());
    

  CurrentBlock ++;
  if (CurrentBlock >= BLOCK_COUNT) CurrentBlock = 0;

// send repeat msg
  if (lncv[CV_REPLAY_MSG_DELAY] != 0)
    if (LastSend[CurrentBlock] != 0)
      if ( (millis() - LastSend[CurrentBlock]) > lncv[CV_REPLAY_MSG_DELAY]) {
          sendSensor1(lncv[CV_ADDR]+CurrentBlock, Tests[CurrentBlock]);
          LastSend[CurrentBlock] = 0;
        }
        
// programmode indication
  ProgrammingModeIdle();
  
//broadcast messages
//  if (millis()-LastBroadcast > BROADCAST_DELAY) {
// sendSensor1(lncv[CV_ADDR]+CurrentBroadcastSensor, Tests[CurrentBroadcastSensor]);
//    CurrentBroadcastSensor ++;
//    if (CurrentBroadcastSensor >= BLOCK_COUNT) CurrentBroadcastSensor = 0;
//    LastBroadcast = millis();
//  }
}


void ProgrammingModeIdle(){
unsigned long CurrentMillis;
  
  if (programmingMode){
     CurrentMillis = millis();
     if (CurrentMillis - LedLastBlink > ledCycle) {
        LedState = !LedState;
        LedLastBlink = CurrentMillis;
        digitalWrite(LED_PIN, LedState);
     }
     if (CurrentMillis - programmingModeStart > PROG_MODE_TIMEOUT) {
        LedState = LOW;
        digitalWrite(LED_PIN, LedState);
        programmingMode = false;
        programmingModeStart = 0;
        Serial.print("End Programing Mode (Auto).\n");
     }
  }
}

// This call-back function is called from LocoNet.processSwitchSensorMessage
// for all Sensor messages
void notifySensor( uint16_t Address, uint8_t State )
{
  Serial.print("Sensor: ");
  Serial.print(Address, DEC);
  Serial.print(" - ");
  Serial.println( State ? "Active" : "Inactive" );
}

int8_t notifyLNCVdiscover( uint16_t & ArtNr, uint16_t & ModuleAddress ){
  Serial.print("Switch Request: ");
  Serial.print(ArtNr, DEC);
  Serial.print(':');
  Serial.print("ModuleAddress:");
  Serial.println(ModuleAddress, DEC);
//  Serial.println(Output ? "On" : "Off");
}

  // This call-back function is called from LocoNet.processSwitchSensorMessage
  // for all Switch Request messages
void notifySwitchRequest( uint16_t Address, uint8_t Output, uint8_t Direction )
{
  Serial.print("Switch Request: ");
  Serial.print(Address, DEC);
  Serial.print(':');
  Serial.print(Direction ? "Closed" : "Thrown");
  Serial.print(" - ");
  Serial.println(Output ? "On" : "Off");
}

  // This call-back function is called from LocoNet.processSwitchSensorMessage
  // for all Switch Report messages
void notifySwitchReport( uint16_t Address, uint8_t Output, uint8_t Direction )
{
  Serial.print("Switch Report: ");
  Serial.print(Address, DEC);
  Serial.print(':');
  Serial.print(Direction ? "Closed" : "Thrown");
  Serial.print(" - ");
  Serial.println(Output ? "On" : "Off");
}

  // This call-back function is called from LocoNet.processSwitchSensorMessage
  // for all Switch State messages
void notifySwitchState( uint16_t Address, uint8_t Output, uint8_t Direction )
{
  Serial.print("Switch State: ");
  Serial.print(Address, DEC);
  Serial.print(':');
  Serial.print(Direction ? "Closed" : "Thrown");
  Serial.print(" - ");
  Serial.println(Output ? "On" : "Off");
}


void sendSensor1 (int addr, boolean state)
{
  byte ret = LocoNet.reportSensor(addr, state);
  Serial.print("SEND addr:");
  Serial.print(addr);
  Serial.print(" state:");
  Serial.print(state);
  Serial.print("ret:");
  Serial.println(ret);
}


void BlockBusy(int Block, unsigned long Time){
  if (!LastOn[Block]) {
      sendSensor1(lncv[CV_ADDR] + Block, true);
      LastSend[Block] = millis();
      Serial.print("ON: ");
      Serial.println(lncv[CV_ADDR] + Block); 
  }      
  LastOn[Block] = Time;    
}

void BlockFree(int Block){
  if (LastOn[Block]) {
      sendSensor1(lncv[CV_ADDR] + Block, false);
      LastSend[Block] = millis();
      Serial.print("OFF: ");
      Serial.println(lncv[CV_ADDR] + Block);
  }
  LastOn[Block] = 0;    
}

void sendAll(){
  for (int i = 0; i < BLOCK_COUNT; i++) {
    sendSensor1(lncv[CV_ADDR] + i, Tests[i]);
    //pinMode(Pins[i], INPUT_PULLUP);
    LastOn[i] = 0;
  }    
    
}

  /**
   * Notifies the code on the reception of a read request.
   * Note that without application knowledge (i.e., art.-nr., module address
   * and "Programming Mode" state), it is not possible to distinguish
   * a read request from a programming start request message.
   */
int8_t notifyLNCVread(uint16_t ArtNr, uint16_t lncvAddress, uint16_t,
    uint16_t & lncvValue) {
  Serial.print("Enter notifyLNCVread(");
  Serial.print(ArtNr, HEX);
  Serial.print(", ");
  Serial.print(lncvAddress, HEX);
  Serial.print(", ");
  Serial.print(", ");
  Serial.print(lncvValue, HEX);
  Serial.print(")");
  // Step 1: Can this be addressed to me?
  // All ReadRequests contain the ARTNR. For starting programming, we do not accept the broadcast address.
  if (programmingMode) {
    if (ArtNr == ARTNR) {
      if (lncvAddress < LNCV_COUNT) {
        lncvValue = lncv[lncvAddress];
        Serial.print(" LNCV Value: ");
        Serial.print(lncvValue);
        Serial.print("\n");
        return LNCV_LACK_OK;
      } else {
        // Invalid LNCV address, request a NAXK
        return LNCV_LACK_ERROR_UNSUPPORTED;
      }
    } else {
      Serial.print("ArtNr invalid.\n");
      return -1;
    }
  } else {
    Serial.print("Ignoring Request.\n");
    return -1;
  }
}

int8_t notifyLNCVprogrammingStart(uint16_t & ArtNr, uint16_t & ModuleAddress) {
  // Enter programming mode. If we already are in programming mode,
  // we simply send a response and nothing else happens.
  Serial.print("notifyLNCVProgrammingStart artn:");
  Serial.print(ArtNr);
  if (ArtNr == ARTNR) {
    Serial.print("artnrOK ");
    if (ModuleAddress == lncv[0]) {
      Serial.print("moduleUNI ENTERING PROGRAMMING MODE\n");
      programmingMode = true;
      programmingModeStart = millis();
      return LNCV_LACK_OK;
    } else if (ModuleAddress == 0xFFFF) {
      Serial.print("moduleBC ENTERING PROGRAMMING MODE\n");
      ModuleAddress = lncv[0];
      return LNCV_LACK_OK;
    }
  }
  else 
  Serial.print("Ignoring Request.\n");
  return -1;
}

  /**
   * Notifies the code on the reception of a write request
   */
int8_t notifyLNCVwrite(uint16_t ArtNr, uint16_t lncvAddress,
    uint16_t lncvValue) {
  Serial.print("notifyLNCVwrite, ");
  //  dumpPacket(ub);
  if (!programmingMode) {
    Serial.print("not in Programming Mode.\n");
    return -1;
  }

  if (ArtNr == ARTNR) {
    Serial.print("Artnr OK, ");

    if (lncvAddress < LNCV_COUNT) {
      
      programmingModeStart = millis();
      
      lncv[lncvAddress] = lncvValue;
      EEPROMSaveCV(lncvAddress);
      //EPROM.writeBlock(sizeof(lncv[0]) * lncvAddress, lncv[lncvAddress]);
      return LNCV_LACK_OK;
    }
    else {
      return LNCV_LACK_ERROR_UNSUPPORTED;
    }

  }
  else {
    Serial.print("Artnr Invalid.\n");
    return -1;
  }
}

void commitLNCVUpdate() {
  Serial.print("Module Address is now: ");
  Serial.print(lncv[CV_ADDR]);
  Serial.print("\n");
}

  /**
   * Notifies the code on the reception of a request to end programming mode
   */
void notifyLNCVprogrammingStop(uint16_t ArtNr, uint16_t ModuleAddress) {
  Serial.print("notifyLNCVprogrammingStop ");
  if (programmingMode) {
    if (ArtNr == ARTNR && ModuleAddress == lncv[0]) {
      programmingMode = false;
      Serial.print("End Programing Mode.\n");
      commitLNCVUpdate();
    }
    else {
      if (ArtNr != ARTNR) {
        Serial.print("Wrong Artnr.\n");
        return;
      }
      if (ModuleAddress != lncv[0]) {
        Serial.print("Wrong Module Address.\n");
        return;
      }
    }
  }
  else {
    Serial.print("Ignoring Request.\n");
  }
}


void dumpPacket(UhlenbrockMsg & ub) {
  Serial.print(" PKT: ");
  Serial.print(ub.command, HEX);
  Serial.print(" ");
  Serial.print(ub.mesg_size, HEX);
  Serial.print(" ");
  Serial.print(ub.SRC, HEX);
  Serial.print(" ");
  Serial.print(ub.DSTL, HEX);
  Serial.print(" ");
  Serial.print(ub.DSTH, HEX);
  Serial.print(" ");
  Serial.print(ub.ReqId, HEX);
  Serial.print(" ");
  Serial.print(ub.PXCT1, HEX);
  Serial.print(" ");
  for (int i(0); i < 8; ++i) {
//     Serial.print(ub.Data[i], HEX);
    Serial.print(" ");
  }
  Serial.write("\n");
}


